#include <Windows.h>
#include <WinIoCtl.h>
#include <stdio.h>

#define BUF_LEN 4096

// Format the Win32 system error code to string
void ErrorMessage(DWORD dwCode);

int wmain(int argc, WCHAR **argv)
{
   HANDLE hVol;
   CHAR Buffer[BUF_LEN];
   USN_JOURNAL_DATA JournalData;
   READ_USN_JOURNAL_DATA ReadData = {0, 0xFFFFFFFF, FALSE, 0, 0};
   PUSN_RECORD UsnRecord;
   DWORD dwBytes;
   DWORD dwRetBytes;
   int i;

   hVol = CreateFile( L"\\\\.\\C:", 
               GENERIC_READ | GENERIC_WRITE, 
               FILE_SHARE_READ | FILE_SHARE_WRITE,
               NULL,
               OPEN_EXISTING,
               0,
               NULL);

   if(hVol == INVALID_HANDLE_VALUE)
   {
      wprintf(L"CreateFile() failed\n");
	  ErrorMessage(GetLastError());
      return 1;
   }

   wprintf(L"CreateFile() is OK!\n");

   if(!DeviceIoControl(hVol, 
          FSCTL_QUERY_USN_JOURNAL, 
          NULL,
          0,
          &JournalData,
          sizeof(JournalData),
          &dwBytes,
          NULL))
   {
      wprintf(L"DeviceIoControl() - Query journal failed\n");
	  ErrorMessage(GetLastError());
      return 1;
   }

   ReadData.UsnJournalID = JournalData.UsnJournalID;

   wprintf(L"Journal ID: %I64x\n", JournalData.UsnJournalID );
   wprintf(L"FirstUsn: %I64x\n\n", JournalData.FirstUsn );

   for(i=0; i<=10; i++)
   {
      memset(Buffer, 0, BUF_LEN);

      if(!DeviceIoControl( hVol, 
            FSCTL_READ_USN_JOURNAL, 
            &ReadData,
            sizeof(ReadData),
            &Buffer,
            BUF_LEN,
            &dwBytes,
            NULL))
      {
         wprintf(L"DeviceIoControl()- Read journal failed\n");
		 ErrorMessage(GetLastError());
         return 1;
      }
	  
	  wprintf(L"DeviceIoControl() is OK!\n");

      dwRetBytes = dwBytes - sizeof(USN);

      // Find the first record
      UsnRecord = (PUSN_RECORD)(((PUCHAR)Buffer) + sizeof(USN));  

      wprintf(L"****************************************\n");

      // This loop could go on for a long time, given the current buffer size.
      while(dwRetBytes > 0)
      {
         wprintf(L"USN: %I64x\n", UsnRecord->Usn );
         wprintf(L"File name: %.*S\n", UsnRecord->FileNameLength/2,UsnRecord->FileName );
         wprintf(L"Reason: %x\n", UsnRecord->Reason );
         wprintf(L"\n" );

         dwRetBytes -= UsnRecord->RecordLength;
         // Find the next record
         UsnRecord = (PUSN_RECORD)(((PCHAR)UsnRecord) + UsnRecord->RecordLength); 
      }
      // Update starting USN for next call
      ReadData.StartUsn = *(USN *)&Buffer; 
   }

   if(CloseHandle(hVol) != 0)
         wprintf(L"CloseHandle() is OK!\n");
   else
   {
         wprintf(L"CloseHandle() failed\n");
         ErrorMessage(GetLastError());
   }
   return 0;
}

void ErrorMessage(DWORD dwCode)
{
    // get the error code...
    DWORD dwErrCode = dwCode;
      DWORD dwNumChar;

    LPWSTR szErrString = NULL;  // will be allocated and filled by FormatMessage

    dwNumChar = FormatMessage( FORMAT_MESSAGE_ALLOCATE_BUFFER |
                 FORMAT_MESSAGE_FROM_SYSTEM, // use windows internal message table
                 0,       // 0 since source is internal message table
                 dwErrCode, // this is the error code number
                 0,       // auto-determine language to use
                 (LPWSTR)&szErrString, // the message
                 0,                 // min size for buffer
                 0 );               // since getting message from system tables
      if(dwNumChar == 0)
            wprintf(L"FormatMessage() failed, error %u\n", GetLastError());
      //else
      //    wprintf(L"FormatMessage() should be fine!\n");

     wprintf(L"Error code %u:\n  %s\n", dwErrCode, szErrString) ;

    // This buffer used by FormatMessage()
    if(LocalFree(szErrString) != NULL)
            wprintf(L"Failed to free up the buffer, error %u\n", GetLastError());
    //else
    //    wprintf(L"Buffer has been freed\n");
  }